@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2010 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -51,10 +51,8 @@
 #define SHUT_RDWR 2
 #endif
 
-#if SSLEAY_VERSION_NUMBER >= 0x0922
 static char *sid_ctx="stunnel SID";
     /* const allowed here */
-#endif
 
 static void do_client(CLI *);
 static void run_client(CLI *);
@@ -69,13 +67,13 @@ static void auth_user(CLI *);
 static int connect_local(CLI *);
 static void make_sockets(CLI *, int [2]);
 static int connect_remote(CLI *);
+#ifdef SO_ORIGINAL_DST
+static int connect_transparent(CLI *);
+#endif /* SO_ORIGINAL_DST */
 static void local_bind(CLI *c);
 static void print_bound_address(CLI *);
 static void reset(int, char *);
 
-int max_clients;
-int max_fds;
-
 /* allocate local data structure for the new thread */
 CLI *alloc_client_session(SERVICE_OPTIONS *opt, int rfd, int wfd) {
     CLI *c;
@@ -98,7 +96,6 @@ void *client(void *arg) {
     stack_info(1); /* initialize */
 #endif
     s_log(LOG_DEBUG, "Service %s started", c->opt->servname);
-#ifndef USE_WIN32
     if(c->opt->option.remote && c->opt->option.program) {
             /* connect and exec options specified together */
             /* -> spawn a local program instead of stdio */
@@ -109,15 +106,7 @@ void *client(void *arg) {
             sleep(1); /* FIXME: not a good idea in ucontext threading */
         }
     } else
-#endif
-    {
-        if(alloc_fd(c->local_rfd.fd))
-            return NULL;
-        if(c->local_wfd.fd!=c->local_rfd.fd)
-            if(alloc_fd(c->local_wfd.fd))
-                return NULL;
         run_client(c);
-    }
     free(c);
 #ifdef DEBUG_STACK_SIZE
     stack_info(0); /* display computed value */
@@ -149,7 +138,7 @@ static void run_client(CLI *c) {
         "Connection %s: %d bytes sent to SSL, %d bytes sent to socket",
          error==1 ? "reset" : "closed", c->ssl_bytes, c->sock_bytes);
 
-        /* cleanup IDENT socket */
+        /* cleanup temporary (e.g. IDENT) socket */
     if(c->fd>=0)
         closesocket(c->fd);
 
@@ -217,7 +206,7 @@ static void init_local(CLI *c) {
 #ifdef USE_WIN32
         if(get_last_socket_error()!=ENOTSOCK) {
 #else
-        if(c->opt->option.transparent || get_last_socket_error()!=ENOTSOCK) {
+        if(c->opt->option.transparent_src || get_last_socket_error()!=ENOTSOCK) {
 #endif
             sockerror("getpeerbyname");
             longjmp(c->err, 1);
@@ -247,7 +236,7 @@ static void init_remote(CLI *c) {
     if(c->opt->source_addr.num)
         memcpy(&c->bind_addr, &c->opt->source_addr, sizeof(SOCKADDR_LIST));
 #ifndef USE_WIN32
-    else if(c->opt->option.transparent)
+    else if(c->opt->option.transparent_src)
         memcpy(&c->bind_addr, &c->peer_addr, sizeof(SOCKADDR_LIST));
 #endif
     else {
@@ -255,16 +244,15 @@ static void init_remote(CLI *c) {
     }
 
     /* setup c->remote_fd, now */
-    if(c->opt->option.remote) {
+    if(c->opt->option.remote)
         c->remote_fd.fd=connect_remote(c);
-    } else /* NOT in remote mode */
+#ifdef SO_ORIGINAL_DST
+    else if(c->opt->option.transparent_dst)
+        c->remote_fd.fd=connect_transparent(c);
+#endif /* SO_ORIGINAL_DST */
+    else /* NOT in remote mode */
         c->remote_fd.fd=connect_local(c);
     c->remote_fd.is_socket=1; /* always! */
-    if(max_fds && c->remote_fd.fd>=max_fds) {
-        s_log(LOG_ERR, "Remote file descriptor out of range (%d>=%d)",
-            c->remote_fd.fd, max_fds);
-        longjmp(c->err, 1);
-    }
     s_log(LOG_DEBUG, "Remote FD=%d initialized", c->remote_fd.fd);
     if(set_socket_options(c->remote_fd.fd, 2)<0)
         longjmp(c->err, 1);
@@ -279,10 +267,8 @@ static void init_ssl(CLI *c) {
         longjmp(c->err, 1);
     }
     SSL_set_ex_data(c->ssl, cli_index, c); /* for callbacks */
-#if SSLEAY_VERSION_NUMBER >= 0x0922
     SSL_set_session_id_context(c->ssl, (unsigned char *)sid_ctx,
         strlen(sid_ctx));
-#endif
     if(c->opt->option.client) {
         if(c->opt->session) {
             enter_critical_section(CRIT_SESSION);
@@ -314,14 +300,19 @@ static void init_ssl(CLI *c) {
     }
 
     while(1) {
-        /* crude workaround for random MT-safety problems in OpenSSL */
-        /* performance penalty is not huge, as it's a non-blocking code */
+#if OPENSSL_VERSION_NUMBER<0x1000002f
+        /* this critical section is a crude workaround for CVE-2010-3864 */
+        /* see http://www.securityfocus.com/bid/44884 for details */
+        /* NOTE: this critical section also covers callbacks (e.g. OCSP) */
         enter_critical_section(CRIT_SSL);
+#endif /* OpenSSL version < 1.0.0b */
         if(c->opt->option.client)
             i=SSL_connect(c->ssl);
         else
             i=SSL_accept(c->ssl);
+#if OPENSSL_VERSION_NUMBER<0x1000002f
         leave_critical_section(CRIT_SSL);
+#endif /* OpenSSL version < 1.0.0b */
         err=SSL_get_error(c->ssl, i);
         if(err==SSL_ERROR_NONE)
             break; /* ok -> done */
@@ -335,12 +326,13 @@ static void init_ssl(CLI *c) {
                 sockerror("init_ssl: s_poll_wait");
                 longjmp(c->err, 1);
             case 0:
-                s_log(LOG_INFO, "init_ssl: s_poll_wait timeout");
+                s_log(LOG_INFO, "init_ssl: s_poll_wait:"
+                    " TIMEOUTbusy exceeded: sending reset");
                 longjmp(c->err, 1);
             case 1:
                 break; /* OK */
             default:
-                s_log(LOG_ERR, "init_ssl: s_poll_wait unknown result");
+                s_log(LOG_ERR, "init_ssl: s_poll_wait: unknown result");
                 longjmp(c->err, 1);
             }
             continue; /* ok -> retry */
@@ -376,61 +368,45 @@ static void init_ssl(CLI *c) {
     }
 }
 
-/****************************** some defines for transfer() */
-/* is socket/SSL open for read/write? */
-#define sock_rd (c->sock_rfd->rd)
-#define sock_wr (c->sock_wfd->wr)
-#define ssl_rd  (c->ssl_rfd->rd)
-#define ssl_wr  (c->ssl_wfd->wr)
-/* NOTE: above defines are related to the logical data stream,
- * not the underlying file descriptors */
-
-/* is socket/SSL ready for read/write? */
-#define sock_can_rd (s_poll_canread(&c->fds, c->sock_rfd->fd))
-#define sock_can_wr (s_poll_canwrite(&c->fds, c->sock_wfd->fd))
-#define ssl_can_rd  (s_poll_canread(&c->fds, c->ssl_rfd->fd))
-#define ssl_can_wr  (s_poll_canwrite(&c->fds, c->ssl_wfd->fd))
-
 /****************************** transfer data */
 static void transfer(CLI *c) {
     int watchdog=0; /* a counter to detect an infinite loop */
-    int error;
-    socklen_t optlen;
-    int num, err, check_SSL_pending;
-    int SSL_shutdown_wants_read=0, SSL_shutdown_wants_write=0;
-    int SSL_write_wants_read=0, SSL_write_wants_write=0;
-    int SSL_read_wants_read=0, SSL_read_wants_write=0;
+    int num, err;
+    /* logical channels (not file descriptors!) open for read or write */
+    int sock_open_rd=1, sock_open_wr=1, ssl_open_rd=1, ssl_open_wr=1;
+    /* awaited conditions on SSL file descriptors */
+    int shutdown_wants_read=0, shutdown_wants_write=0;
+    int read_wants_read, read_wants_write=0;
+    int write_wants_read=0, write_wants_write;
+    /* actual conditions on file descriptors */
+    int sock_can_rd, sock_can_wr, ssl_can_rd, ssl_can_wr;
 
     c->sock_ptr=c->ssl_ptr=0;
-    sock_rd=sock_wr=ssl_rd=ssl_wr=1;
 
-    do { /* main loop */
-        /* set flag to try and read any buffered SSL data
-         * if we made room in the buffer by writing to the socket */
-        check_SSL_pending=0;
-
-        SSL_read_wants_read=
-            ssl_rd && c->ssl_ptr<BUFFSIZE && !SSL_read_wants_write;
-        SSL_write_wants_write=
-            ssl_wr && c->sock_ptr && !SSL_write_wants_read;
+    do { /* main loop of client data transfer */
+        /****************************** initialize *_wants_* */
+        read_wants_read=
+            ssl_open_rd && c->ssl_ptr<BUFFSIZE && !read_wants_write;
+        write_wants_write=
+            ssl_open_wr && c->sock_ptr && !write_wants_read;
 
         /****************************** setup c->fds structure */
         s_poll_init(&c->fds); /* initialize the structure */
-        if(sock_rd && c->sock_ptr<BUFFSIZE)
-            s_poll_add(&c->fds, c->sock_rfd->fd, 1, 0);
-        if(SSL_read_wants_read ||
-                SSL_write_wants_read ||
-                SSL_shutdown_wants_read)
+        /* for plain socket open data strem = open file descriptor */
+        /* make sure to add each open socket to receive exceptions! */
+        if(sock_open_rd)
+            s_poll_add(&c->fds, c->sock_rfd->fd, c->sock_ptr<BUFFSIZE, 0);
+        if(sock_open_wr)
+            s_poll_add(&c->fds, c->sock_wfd->fd, 0, c->ssl_ptr);
+        /* for SSL assume that sockets are open if there any pending requests */
+        if(read_wants_read || write_wants_read || shutdown_wants_read)
             s_poll_add(&c->fds, c->ssl_rfd->fd, 1, 0);
-        if(sock_wr && c->ssl_ptr)
-            s_poll_add(&c->fds, c->sock_wfd->fd, 0, 1);
-        if(SSL_read_wants_write ||
-                SSL_write_wants_write ||
-                SSL_shutdown_wants_write)
+        if(read_wants_write || write_wants_write || shutdown_wants_write)
             s_poll_add(&c->fds, c->ssl_wfd->fd, 0, 1);
 
         /****************************** wait for an event */
-        err=s_poll_wait(&c->fds, (sock_rd && ssl_rd) /* both peers open */ ||
+        err=s_poll_wait(&c->fds,
+            (sock_open_rd && ssl_open_rd) /* both peers open */ ||
             c->ssl_ptr /* data buffered to write to socket */ ||
             c->sock_ptr /* data buffered to write to SSL */ ?
             c->opt->timeout_idle : c->opt->timeout_close, 0);
@@ -439,41 +415,88 @@ static void transfer(CLI *c) {
             sockerror("transfer: s_poll_wait");
             longjmp(c->err, 1);
         case 0: /* timeout */
-            if((sock_rd && ssl_rd) || c->ssl_ptr || c->sock_ptr) {
-                s_log(LOG_INFO, "s_poll_wait timeout: connection reset");
+            if((sock_open_rd && ssl_open_rd) || c->ssl_ptr || c->sock_ptr) {
+                s_log(LOG_INFO, "transfer: s_poll_wait:"
+                    " TIMEOUTidle exceeded: sending reset");
                 longjmp(c->err, 1);
             } else { /* already closing connection */
-                s_log(LOG_INFO, "s_poll_wait timeout: connection close");
+                s_log(LOG_ERR, "transfer: s_poll_wait:"
+                    " TIMEOUTclose exceeded: closing");
                 return; /* OK */
             }
         }
+
+        /****************************** check for errors on sockets */
+        err=s_poll_error(&c->fds, c->sock_rfd->fd);
+        if(err) {
+            s_log(LOG_NOTICE,
+                "Error detected on socket (read) file descriptor: %s (%d)",
+                s_strerror(err), err);
+            longjmp(c->err, 1);
+        }
+        if(c->sock_wfd->fd != c->sock_rfd->fd) { /* performance optimization */
+            err=s_poll_error(&c->fds, c->sock_wfd->fd);
+            if(err) {
+                s_log(LOG_NOTICE,
+                    "Error detected on socket write file descriptor: %s (%d)",
+                    s_strerror(err), err);
+                longjmp(c->err, 1);
+            }
+        }
+        err=s_poll_error(&c->fds, c->ssl_rfd->fd);
+        if(err) {
+            s_log(LOG_NOTICE,
+                "Error detected on SSL (read) file descriptor: %s (%d)",
+                s_strerror(err), err);
+            longjmp(c->err, 1);
+        }
+        if(c->ssl_wfd->fd != c->ssl_rfd->fd) { /* performance optimization */
+            err=s_poll_error(&c->fds, c->ssl_wfd->fd);
+            if(err) {
+                s_log(LOG_NOTICE,
+                    "Error detected on SSL write file descriptor: %s (%d)",
+                    s_strerror(err), err);
+                longjmp(c->err, 1);
+            }
+        }
+
+        /****************************** retrieve results from c->fds */
+        sock_can_rd=s_poll_canread(&c->fds, c->sock_rfd->fd);
+        sock_can_wr=s_poll_canwrite(&c->fds, c->sock_wfd->fd);
+        ssl_can_rd=s_poll_canread(&c->fds, c->ssl_rfd->fd);
+        ssl_can_wr=s_poll_canwrite(&c->fds, c->ssl_wfd->fd);
+
+        /****************************** checks for internal failures */
+        /* please report any internal errors to stunnel-users mailing list */
         if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
             s_log(LOG_ERR, "INTERNAL ERROR: "
                 "s_poll_wait returned %d, but no descriptor is ready", err);
             longjmp(c->err, 1);
         }
-        if(!sock_rd && sock_can_rd) {
-            optlen=sizeof error;
-            if(getsockopt(c->sock_rfd->fd, SOL_SOCKET, SO_ERROR,
-                    (void *)&error, &optlen))
-                error=get_last_socket_error(); /* failed -> ask why */
-            if(error) { /* really an error? */
-                s_log(LOG_ERR, "Closed socket ready to read: %s (%d)",
-                    my_strerror(error), error);
+        /* these checks should no longer be needed */
+        /* I'm going to remove them soon */
+        if(!sock_open_rd && sock_can_rd) {
+            err=get_socket_error(c->sock_rfd->fd);
+            if(err) { /* really an error? */
+                s_log(LOG_ERR, "INTERNAL ERROR: "
+                    "Closed socket ready to read: %s (%d)",
+                    s_strerror(err), err);
                 longjmp(c->err, 1);
             }
             if(c->ssl_ptr) { /* anything left to write */
-                s_log(LOG_ERR, "Closed socket ready to read - reset");
+                s_log(LOG_ERR, "INTERNAL ERROR: "
+                    "Closed socket ready to read: sending reset");
                 longjmp(c->err, 1);
             }
-            s_log(LOG_INFO, "Closed socket ready to read - write close");
-            sock_wr=0; /* no further write allowed */
+            s_log(LOG_ERR, "INTERNAL ERROR: "
+                "Closed socket ready to read: write close");
+            sock_open_wr=0; /* no further write allowed */
             shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
         }
 
         /****************************** send SSL close_notify message */
-        if(SSL_shutdown_wants_read || SSL_shutdown_wants_write) {
-            SSL_shutdown_wants_read=SSL_shutdown_wants_write=0;
+        if(shutdown_wants_read || shutdown_wants_write) {
+            shutdown_wants_read=shutdown_wants_write=0;
             num=SSL_shutdown(c->ssl); /* send close_notify */
             if(num<0) /* -1 - not completed */
                 err=SSL_get_error(c->ssl, num);
@@ -485,11 +508,11 @@ static void transfer(CLI *c) {
                 break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_shutdown returned WANT_WRITE: retrying");
-                SSL_shutdown_wants_write=1;
+                shutdown_wants_write=1;
                 break;
             case SSL_ERROR_WANT_READ:
                 s_log(LOG_DEBUG, "SSL_shutdown returned WANT_READ: retrying");
-                SSL_shutdown_wants_read=1;
+                shutdown_wants_read=1;
                 break;
             case SSL_ERROR_SYSCALL: /* socket error */
                 parse_socket_error(c, "SSL_shutdown");
@@ -503,8 +526,26 @@ static void transfer(CLI *c) {
             }
         }
 
+        /****************************** read from socket */
+        if(sock_open_rd && sock_can_rd) {
+            num=readsocket(c->sock_rfd->fd,
+                c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
+            switch(num) {
+            case -1:
+                parse_socket_error(c, "readsocket");
+                break;
+            case 0: /* close */
+                s_log(LOG_DEBUG, "Socket closed on read");
+                sock_open_rd=0;
+                break;
+            default:
+                c->sock_ptr+=num;
+                watchdog=0; /* reset watchdog */
+            }
+        }
+
         /****************************** write to socket */
-        if(sock_wr && sock_can_wr) {
+        if(sock_open_wr && sock_can_wr) {
             num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
             switch(num) {
             case -1: /* error */
@@ -515,88 +556,25 @@ static void transfer(CLI *c) {
                 break;
             default:
                 memmove(c->ssl_buff, c->ssl_buff+num, c->ssl_ptr-num);
-                if(c->ssl_ptr==BUFFSIZE) /* buffer was previously full */
-                    check_SSL_pending=1; /* check for data buffered by SSL */
                 c->ssl_ptr-=num;
                 c->sock_bytes+=num;
                 watchdog=0; /* reset watchdog */
             }
         }
 
-        /****************************** write to SSL */
-        if((SSL_write_wants_read && ssl_can_rd) ||
-                (SSL_write_wants_write && ssl_can_wr)) {
-            SSL_write_wants_read=0;
-            num=SSL_write(c->ssl, c->sock_buff, c->sock_ptr);
-            switch(err=SSL_get_error(c->ssl, num)) {
-            case SSL_ERROR_NONE:
-                memmove(c->sock_buff, c->sock_buff+num, c->sock_ptr-num);
-                c->sock_ptr-=num;
-                c->ssl_bytes+=num;
-                watchdog=0; /* reset watchdog */
-                break;
-            case SSL_ERROR_WANT_WRITE: /* nothing unexpected */
-                break;
-            case SSL_ERROR_WANT_READ:
-                s_log(LOG_DEBUG, "SSL_write returned WANT_READ: retrying");
-                SSL_write_wants_read=1;
-                break;
-            case SSL_ERROR_WANT_X509_LOOKUP:
-                s_log(LOG_DEBUG,
-                    "SSL_write returned WANT_X509_LOOKUP: retrying");
-                break;
-            case SSL_ERROR_SYSCALL: /* socket error */
-                if(!num) { /* EOF */
-                    if(c->sock_ptr) {
-                        s_log(LOG_ERR,
-                            "SSL socket closed on SSL_write "
-                                "with %d byte(s) in buffer",
-                            c->sock_ptr);
-                        longjmp(c->err, 1); /* reset the socket */
-                    }
-                    s_log(LOG_DEBUG, "SSL socket closed on SSL_write");
-                    ssl_rd=ssl_wr=0; /* buggy peer: no close_notify */
-                } else
-                    parse_socket_error(c, "SSL_write");
-                break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
-                s_log(LOG_DEBUG, "SSL closed on SSL_write");
-                ssl_rd=0;
-                if(!strcmp(SSL_get_version(c->ssl), "SSLv2"))
-                    ssl_wr=0;
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_write");
-                longjmp(c->err, 1);
-            default:
-                s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
-                longjmp(c->err, 1);
-            }
-        }
-
-        /****************************** read from socket */
-        if(sock_rd && sock_can_rd) {
-            num=readsocket(c->sock_rfd->fd,
-                c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
-            switch(num) {
-            case -1:
-                parse_socket_error(c, "readsocket");
-                break;
-            case 0: /* close */
-                s_log(LOG_DEBUG, "Socket closed on read");
-                sock_rd=0;
-                break;
-            default:
-                c->sock_ptr+=num;
-                watchdog=0; /* reset watchdog */
-            }
-        }
+        /****************************** update *_wants_* based on new *_ptr */
+        /* this update is also required for SSL_pending() to be used */
+        read_wants_read=
+            ssl_open_rd && c->ssl_ptr<BUFFSIZE && !read_wants_write;
+        write_wants_write=
+            ssl_open_wr && c->sock_ptr && !write_wants_read;
 
         /****************************** read from SSL */
-        if((SSL_read_wants_read && ssl_can_rd) ||
-                (SSL_read_wants_write && ssl_can_wr) ||
-                (check_SSL_pending && SSL_pending(c->ssl))) {
-            SSL_read_wants_write=0;
+        if((read_wants_read && (ssl_can_rd || SSL_pending(c->ssl))) ||
+                /* it may be possible to read some pending data after
+                 * writesocket() above made some room in c->ssl_buff */
+                (read_wants_write && ssl_can_wr)) {
+            read_wants_write=0;
             num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, BUFFSIZE-c->ssl_ptr);
             switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
@@ -605,7 +583,7 @@ static void transfer(CLI *c) {
                 break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_read returned WANT_WRITE: retrying");
-                SSL_read_wants_write=1;
+                read_wants_write=1;
                 break;
             case SSL_ERROR_WANT_READ: /* nothing unexpected */
                 break;
@@ -623,15 +601,15 @@ static void transfer(CLI *c) {
                         longjmp(c->err, 1); /* reset the socket */
                     }
                     s_log(LOG_DEBUG, "SSL socket closed on SSL_read");
-                    ssl_rd=ssl_wr=0; /* buggy peer: no close_notify */
+                    ssl_open_rd=ssl_open_wr=0; /* buggy peer: no close_notify */
                 } else
                     parse_socket_error(c, "SSL_read");
                 break;
             case SSL_ERROR_ZERO_RETURN: /* close_notify received */
                 s_log(LOG_DEBUG, "SSL closed on SSL_read");
-                ssl_rd=0;
+                ssl_open_rd=0;
                 if(!strcmp(SSL_get_version(c->ssl), "SSLv2"))
-                    ssl_wr=0;
+                    ssl_open_wr=0;
                 break;
             case SSL_ERROR_SSL:
                 sslerror("SSL_read");
@@ -642,23 +620,74 @@ static void transfer(CLI *c) {
             }
         }
 
+        /****************************** write to SSL */
+        if((write_wants_read && ssl_can_rd) ||
+                (write_wants_write && ssl_can_wr)) {
+            write_wants_read=0;
+            num=SSL_write(c->ssl, c->sock_buff, c->sock_ptr);
+            switch(err=SSL_get_error(c->ssl, num)) {
+            case SSL_ERROR_NONE:
+                memmove(c->sock_buff, c->sock_buff+num, c->sock_ptr-num);
+                c->sock_ptr-=num;
+                c->ssl_bytes+=num;
+                watchdog=0; /* reset watchdog */
+                break;
+            case SSL_ERROR_WANT_WRITE: /* nothing unexpected */
+                break;
+            case SSL_ERROR_WANT_READ:
+                s_log(LOG_DEBUG, "SSL_write returned WANT_READ: retrying");
+                write_wants_read=1;
+                break;
+            case SSL_ERROR_WANT_X509_LOOKUP:
+                s_log(LOG_DEBUG,
+                    "SSL_write returned WANT_X509_LOOKUP: retrying");
+                break;
+            case SSL_ERROR_SYSCALL: /* socket error */
+                if(!num) { /* EOF */
+                    if(c->sock_ptr) {
+                        s_log(LOG_ERR,
+                            "SSL socket closed on SSL_write "
+                                "with %d byte(s) in buffer",
+                            c->sock_ptr);
+                        longjmp(c->err, 1); /* reset the socket */
+                    }
+                    s_log(LOG_DEBUG, "SSL socket closed on SSL_write");
+                    ssl_open_rd=ssl_open_wr=0; /* buggy peer: no close_notify */
+                } else
+                    parse_socket_error(c, "SSL_write");
+                break;
+            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
+                s_log(LOG_DEBUG, "SSL closed on SSL_write");
+                ssl_open_rd=0;
+                if(!strcmp(SSL_get_version(c->ssl), "SSLv2"))
+                    ssl_open_wr=0;
+                break;
+            case SSL_ERROR_SSL:
+                sslerror("SSL_write");
+                longjmp(c->err, 1);
+            default:
+                s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
+                longjmp(c->err, 1);
+            }
+        }
+
         /****************************** check write shutdown conditions */
-        if(sock_wr && !ssl_rd && !c->ssl_ptr) {
-            s_log(LOG_DEBUG, "Socket write shutdown");
-            sock_wr=0; /* no further write allowed */
+        if(sock_open_wr && !ssl_open_rd && !c->ssl_ptr) {
+            s_log(LOG_DEBUG, "Sending socket write shutdown");
+            sock_open_wr=0; /* no further write allowed */
             shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
         }
-        if(ssl_wr && !sock_rd && !c->sock_ptr) {
-            s_log(LOG_DEBUG, "SSL write shutdown");
-            ssl_wr=0; /* no further write allowed */
+        if(ssl_open_wr && !sock_open_rd && !c->sock_ptr) {
+            s_log(LOG_DEBUG, "Sending SSL write shutdown");
+            ssl_open_wr=0; /* no further write allowed */
             if(strcmp(SSL_get_version(c->ssl), "SSLv2")) { /* SSLv3, TLSv1 */
-                SSL_shutdown_wants_write=1; /* initiate close_notify */
+                shutdown_wants_write=1; /* initiate close_notify */
             } else { /* no alerts in SSLv2 including close_notify alert */
                 shutdown(c->sock_rfd->fd, SHUT_RD); /* notify the kernel */
                 shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
                 SSL_set_shutdown(c->ssl, /* notify the OpenSSL library */
                     SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                ssl_rd=0; /* no further read allowed */
+                ssl_open_rd=0; /* no further read allowed */
             }
         }
 
@@ -668,31 +697,33 @@ static void transfer(CLI *c) {
                 "transfer() loop executes not transferring any data");
             s_log(LOG_ERR,
                 "please report the problem to Michal.Trojnara@mirt.net");
-            s_log(LOG_ERR, "protocol=%s, check_SSL_pending=%s",
-                SSL_get_version(c->ssl), check_SSL_pending ? "yes" : "no");
-            s_log(LOG_ERR, "socket open: rd=%s wr=%s, ssl open: rd=%s wr=%s",
-                sock_rd ? "yes" : "no", sock_wr ? "yes" : "no",
-                ssl_rd ? "yes" : "no", ssl_wr ? "yes" : "no");
-            s_log(LOG_ERR, "socket ready: rd=%s wr=%s, ssl ready: rd=%s wr=%s",
-                sock_can_rd ? "yes" : "no", sock_can_wr ? "yes" : "no",
-                ssl_can_rd ? "yes" : "no", ssl_can_wr ? "yes" : "no");
-            s_log(LOG_ERR,
-                "wants: SSL_read rd=%s wr=%s, "
-                "SSL_write rd=%s wr=%s, "
-                "SSL_shutdown rd=%s wr=%s",
-                SSL_read_wants_read ? "yes" : "no",
-                SSL_read_wants_write ? "yes" : "no",
-                SSL_write_wants_read ? "yes" : "no",
-                SSL_write_wants_write ? "yes" : "no",
-                SSL_shutdown_wants_read ? "yes" : "no",
-                SSL_shutdown_wants_write ? "yes" : "no");
+            stunnel_info(LOG_ERR);
+            s_log(LOG_ERR, "protocol=%s, SSL_pending=%d",
+                SSL_get_version(c->ssl), SSL_pending(c->ssl));
+            s_log(LOG_ERR, "sock_open_rd=%s, sock_open_wr=%s, "
+                "ssl_open_rd=%s, ssl_open_wr=%s",
+                sock_open_rd ? "Y" : "n", sock_open_wr ? "Y" : "n",
+                ssl_open_rd ? "Y" : "n", ssl_open_wr ? "Y" : "n");
+            s_log(LOG_ERR, "sock_can_rd=%s,  sock_can_wr=%s,  "
+                "ssl_can_rd=%s,  ssl_can_wr=%s",
+                sock_can_rd ? "Y" : "n", sock_can_wr ? "Y" : "n",
+                ssl_can_rd ? "Y" : "n", ssl_can_wr ? "Y" : "n");
+            s_log(LOG_ERR, "read_wants_read=%s,     read_wants_write=%s",
+                read_wants_read ? "Y" : "n",
+                read_wants_write ? "Y" : "n");
+            s_log(LOG_ERR, "write_wants_read=%s,    write_wants_write=%s",
+                write_wants_read ? "Y" : "n",
+                write_wants_write ? "Y" : "n");
+            s_log(LOG_ERR, "shutdown_wants_read=%s, shutdown_wants_write=%s",
+                shutdown_wants_read ? "Y" : "n",
+                shutdown_wants_write ? "Y" : "n");
             s_log(LOG_ERR, "socket input buffer: %d byte(s), "
                 "ssl input buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
             longjmp(c->err, 1);
         }
 
-    } while(sock_wr || ssl_wr ||
-            SSL_shutdown_wants_read || SSL_shutdown_wants_write);
+    } while(sock_open_wr || ssl_open_wr ||
+        shutdown_wants_read || shutdown_wants_write);
 }
 
 static void parse_socket_error(CLI *c, const char *text) {
@@ -718,10 +749,6 @@ static void parse_socket_error(CLI *c, const char *text) {
 }
 
 static void print_cipher(CLI *c) { /* print negotiated cipher */
-#if SSLEAY_VERSION_NUMBER <= 0x0800
-    s_log(LOG_INFO, "Service %s opened with SSLv%d, cipher %s",
-        c->opt->servname, ssl->session->ssl_version, SSL_get_cipher(c->ssl));
-#else
     SSL_CIPHER *cipher;
     char buf[STRLEN], *i, *j;
 
@@ -742,7 +769,6 @@ static void print_cipher(CLI *c) { /* print negotiated cipher */
         }
     } while(*i++);
     s_log(LOG_INFO, "Negotiated ciphers: %s", buf);
-#endif
 }
 
 static void auth_user(CLI *c) {
@@ -754,12 +780,9 @@ static void auth_user(CLI *c) {
 
     if(!c->opt->username)
         return; /* -u option not specified */
-    if((c->fd=
-            socket(c->peer_addr.addr[0].sa.sa_family, SOCK_STREAM, 0))<0) {
-        sockerror("socket (auth_user)");
-        longjmp(c->err, 1);
-    }
-    if(alloc_fd(c->fd))
+    c->fd=s_socket(c->peer_addr.addr[0].sa.sa_family, SOCK_STREAM,
+        0, 1, "socket (auth_user)");
+    if(c->fd<0)
         longjmp(c->err, 1);
     memcpy(&ident, &c->peer_addr.addr[0], sizeof ident);
 #ifndef _WIN32_WCE
@@ -793,36 +816,43 @@ static void auth_user(CLI *c) {
     s_log(LOG_INFO, "IDENT authentication passed");
 }
 
-#ifdef USE_WIN32
+#if defined(_WIN32_WCE) || defined(__vms)
+
+static int connect_local(CLI *c) { /* spawn local process */
+    s_log(LOG_ERR, "Local mode is not supported on this platform");
+    longjmp(c->err, 1);
+    return -1; /* some C compilers require a return value */
+}
+
+#elif defined(USE_WIN32)
 
 static int connect_local(CLI *c) { /* spawn local process */
     int fd[2];
     STARTUPINFO si;
     PROCESS_INFORMATION pi;
+    LPTSTR execname_l, execargs_l;
 
     make_sockets(c, fd);
-    ZeroMemory(&si, sizeof si);
+    memset(&si, 0, sizeof si);
     si.cb=sizeof si;
     si.wShowWindow=SW_HIDE;
     si.dwFlags=STARTF_USESHOWWINDOW|STARTF_USESTDHANDLES;
     si.hStdInput=si.hStdOutput=si.hStdError=(HANDLE)fd[1];
-    ZeroMemory(&pi, sizeof pi);
-    CreateProcess(c->opt->execname, c->opt->execargs, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
+    memset(&pi, 0, sizeof pi);
+
+    execname_l=str2tstr(c->opt->execname);
+    execargs_l=str2tstr(c->opt->execargs);
+    CreateProcess(execname_l, execargs_l, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
+    free(execname_l);
+    free(execargs_l);
+
     closesocket(fd[1]);
     CloseHandle(pi.hProcess);
     CloseHandle(pi.hThread);
     return fd[0];
 }
 
-#elif defined(__vms)
-
-static int connect_local(CLI *c) { /* spawn local process */
-    s_log(LOG_ERR, "Local mode not supported on OpenVMS");
-    longjmp(c->err, 1);
-    return -1; /* some C compilers require a return value */
-}
-
-#else /* not USE_WIN32 or __vms */
+#else /* standard Unix version */
 
 static int connect_local(CLI *c) { /* spawn local process */
     char env[3][STRLEN], name[STRLEN], *portname;
@@ -835,11 +865,12 @@ static int connect_local(CLI *c) { /* spawn local process */
     if(c->opt->option.pty) {
         char tty[STRLEN];
 
-        if(pty_allocate(fd, fd+1, tty, STRLEN))
+        if(pty_allocate(fd, fd+1, tty))
             longjmp(c->err, 1);
         s_log(LOG_DEBUG, "TTY=%s allocated", tty);
     } else
         make_sockets(c, fd);
+
     pid=fork();
     c->pid=(unsigned long)pid;
     switch(pid) {
@@ -850,18 +881,19 @@ static int connect_local(CLI *c) { /* spawn local process */
         longjmp(c->err, 1);
     case  0:    /* child */
         closesocket(fd[0]);
+        /* dup2() does not copy FD_CLOEXEC flag */
         dup2(fd[1], 0);
         dup2(fd[1], 1);
         if(!global_options.option.foreground)
             dup2(fd[1], 2);
-        closesocket(fd[1]);
+        closesocket(fd[1]); /* not really needed due to FD_CLOEXEC */
         safecopy(env[0], "REMOTE_HOST=");
         safeconcat(env[0], c->accepted_address);
         portname=strrchr(env[0], ':');
         if(portname) /* strip the port name */
             *portname='\0';
         putenv(env[0]);
-        if(c->opt->option.transparent) {
+        if(c->opt->option.transparent_src) {
             putenv("LD_PRELOAD=" LIBDIR "/libstunnel.so");
             /* for Tru64 _RLD_LIST is used instead */
             putenv("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
@@ -889,16 +921,11 @@ static int connect_local(CLI *c) { /* spawn local process */
         execvp(c->opt->execname, c->opt->execargs);
         ioerror(c->opt->execname); /* execv failed */
         _exit(1);
-    default:
-        break;
+    default: /* parent */
+        s_log(LOG_INFO, "Local mode child started (PID=%lu)", c->pid);
+        closesocket(fd[1]);
+        return fd[0];
     }
-    /* parent */
-    s_log(LOG_INFO, "Local mode child started (PID=%lu)", c->pid);
-    closesocket(fd[1]);
-#ifdef FD_CLOEXEC
-    fcntl(fd[0], F_SETFD, FD_CLOEXEC);
-#endif
-    return fd[0];
 }
 
 #endif /* not USE_WIN32 or __vms */
@@ -909,14 +936,12 @@ static void make_sockets(CLI *c, int fd[2]) { /* make a pair of connected socket
     socklen_t addrlen;
     int s; /* temporary socket awaiting for connection */
 
-    if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("socket#1");
+    s=s_socket(AF_INET, SOCK_STREAM, 0, 1, "socket#1");
+    if(s<0)
         longjmp(c->err, 1);
-    }
-    if((fd[1]=socket(AF_INET, SOCK_STREAM, 0))<0) {
-        sockerror("socket#2");
+    fd[1]=s_socket(AF_INET, SOCK_STREAM, 0, 1, "socket#2");
+    if(fd[1]<0)
         longjmp(c->err, 1);
-    }
     addrlen=sizeof addr;
     memset(&addr, 0, addrlen);
     addr.in.sin_family=AF_INET;
@@ -938,20 +963,16 @@ static void make_sockets(CLI *c, int fd[2]) { /* make a pair of connected socket
         sockerror("connect");
         longjmp(c->err, 1);
     }
-    if((fd[0]=accept(s, &addr.sa, &addrlen))<0) {
-        sockerror("accept");
+    if((fd[0]=s_accept(s, &addr.sa, &addrlen, 1, "accept"))<0)
         longjmp(c->err, 1);
-    }
     closesocket(s); /* don't care about the result */
 #else
-    if(socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) {
-        sockerror("socketpair");
+    if(s_socketpair(AF_UNIX, SOCK_STREAM, 0, fd, 1, "socketpair"))
         longjmp(c->err, 1);
-    }
 #endif
 }
 
-static int connect_remote(CLI *c) { /* connect to remote host */
+static int connect_remote(CLI *c) { /* connect remote host */
     SOCKADDR_UNION addr;
     SOCKADDR_LIST resolved_list, *address_list;
     int fd, ind_try, ind_cur;
@@ -979,11 +1000,8 @@ static int connect_remote(CLI *c) { /* connect to remote host */
         }
         memcpy(&addr, address_list->addr+ind_cur, sizeof addr);
 
-        if((c->fd=socket(addr.sa.sa_family, SOCK_STREAM, 0))<0) {
-            sockerror("remote socket");
-            longjmp(c->err, 1);
-        }
-        if(alloc_fd(c->fd))
+        c->fd=s_socket(addr.sa.sa_family, SOCK_STREAM, 0, 1, "remote socket");
+        if(c->fd<0)
             longjmp(c->err, 1);
 
         if(c->bind_addr.num) /* explicit local bind or transparent proxy */
@@ -1003,20 +1021,72 @@ static int connect_remote(CLI *c) { /* connect to remote host */
     return -1; /* some C compilers require a return value */
 }
 
+#ifdef SO_ORIGINAL_DST
+static int connect_transparent(CLI *c) { /* connect the original dst */
+    SOCKADDR_UNION addr;
+    socklen_t addrlen=sizeof addr;
+    int retval;
+
+    if(getsockopt(c->local_rfd.fd, SOL_IP, SO_ORIGINAL_DST,
+            &addr, &addrlen)) {
+        sockerror("setsockopt SO_ORIGINAL_DST");
+        longjmp(c->err, 1);
+    }
+    c->fd=s_socket(addr.sa.sa_family, SOCK_STREAM, 0, 1, "remote socket");
+    if(c->fd<0)
+        longjmp(c->err, 1);
+    if(c->bind_addr.num) /* explicit local bind or transparent proxy */
+        local_bind(c);
+    if(connect_blocking(c, &addr, addr_len(addr)))
+        longjmp(c->err, 1); /* socket closed on cleanup */
+    print_bound_address(c);
+    retval=c->fd;
+    c->fd=-1;
+    return retval; /* success! */
+}
+#endif /* SO_ORIGINAL_DST */
+
 static void local_bind(CLI *c) {
     SOCKADDR_UNION addr;
+    int on;
 
-#ifdef IP_TRANSPARENT
-    int on=1;
-    if(c->opt->option.transparent) {
+    on=1;
+    memcpy(&addr, &c->bind_addr.addr[0], sizeof addr);
+
+#if defined(USE_WIN32)
+    /* do nothing */
+#elif defined(IP_TRANSPARENT)
+    /* non-local bind on Linux */
+    if(c->opt->option.transparent_src) {
         if(setsockopt(c->fd, SOL_IP, IP_TRANSPARENT, &on, sizeof on))
             sockerror("setsockopt IP_TRANSPARENT");
         /* ignore the error to retain Linux 2.2 compatibility */
         /* the error will be handled by bind(), anyway */
     }
-#endif /* IP_TRANSPARENT */
+#elif defined(IP_BINDANY) && defined(IPV6_BINDANY)
+    /* non-local bind on FreeBSD */
+    if(c->opt->option.transparent_src) {
+        if(addr.sa.sa_family==AF_INET) { /* IPv4 */
+            if(setsockopt(c->fd, IPPROTO_IP, IP_BINDANY, &on, sizeof on)) {
+                sockerror("setsockopt IP_BINDANY");
+                longjmp(c->err, 1);
+            }
+        } else { /* IPv6 */
+            if(setsockopt(c->fd, IPPROTO_IPV6, IPV6_BINDANY, &on, sizeof on)) {
+                sockerror("setsockopt IPV6_BINDANY");
+                longjmp(c->err, 1);
+            }
+        }
+    }
+#else
+    /* unsupported platform */
+    if(c->opt->option.transparent_src) {
+        s_log(LOG_ERR, "Transparent proxy in remote mode is not supported"
+            " on this platform");
+        longjmp(c->err, 1);
+    }
+#endif
 
-    memcpy(&addr, &c->bind_addr.addr[0], sizeof addr);
     if(ntohs(addr.in.sin_port)>=1024) { /* security check */
         if(!bind(c->fd, &addr.sa, addr_len(addr))) {
             s_log(LOG_INFO, "local_bind succeeded on the original port");
@@ -1024,7 +1094,7 @@ static void local_bind(CLI *c) {
         }
         if(get_last_socket_error()!=EADDRINUSE
 #ifndef USE_WIN32
-                || !c->opt->option.transparent
+                || !c->opt->option.transparent_src
 #endif /* USE_WIN32 */
                 ) {
             sockerror("local_bind (original port)");
