@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2010 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -72,7 +72,7 @@
  * the buffer must be able to hold at least 64 characters
  */
 
-int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf, int namebuflen) {
+int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf) {
 #if defined(HAVE_OPENPTY) || defined(BSD4_4) && !defined(__INNOTEK_LIBC__)
     /* openpty(3) exists in OSF/1 and some other os'es */
     char buf[64];
@@ -195,7 +195,7 @@ int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf, int namebuflen) {
         if(*ptyfd<0)
             continue;
 #ifdef HAVE_SNPRINTF
-        snprintf(namebuf, namebuflen,
+        snprintf(namebuf, STRLEN,
 #else
         sprintf(namebuf,
 #endif
